import json
import os


OP_DEV_CONFIG = json.loads(os.environ.get('OP_DEV_CONFIG', '{}'))

OP_DEV_METADATA = json.loads(os.environ.get('OP_DEV_METADATA', '{}')) or { 'domain': None, 'pkg_id': None, 'commit_id': None, 'config_id': None }
