import setuptools
import sys

assert sys.version_info >= (3, 6)

setuptools.setup(
  name='op-service',
  version='0.1.2',
  author='Alex Exarhos',
  author_email='AlexExarhos@gmail.com',
  packages=[
    'op_service'
  ],
  install_requires=[
    'iso8601 >= 0.1.11',
    'op-alarm >= 0.1.0',
    'op-distributed-task >= 0.1.0',
    'op-service-client >= 0.1.0',
    'op-storage >= 0.1.0',
    'PyJWT >= 1.4.2',
    'python-multipart >= 0.0.5',
    'starlette >= 0.13.0'
  ]
)
